package main

import (
	_ "expvar"
	"flag"
	"fmt"
	"math/rand"
	"net/http"
	_ "net/http/pprof"
	"os"
	"os/signal"
	"runtime"
	"runtime/debug"
	"sync/atomic"
	"syscall"
	"time"
)

var (
	pprofPort    = flag.String("pprof-port", ":3000", "port to run pprof on")
	startBufs    = flag.Int("start-bufs", 0, "bufs to allocate and discard up front")
	startBatches = flag.Int("start-batches", 0, "number of times to allocate and discard start-bufs; if non-zero, start-bufs will be generated at bufs-per-sec")
	bps          = flag.Int("bufs-per-sec", 10000, "number of 4KiB-12KiB buffers to to generate per second")

	numCreated   int64 // an atomic int incremented in the creator goroutine
	bytesCreated int64 // an atomic int solely to avoid optimizing new slices out
)

func creator() {
	for range time.Tick(time.Second / time.Duration(*bps)) {
		newBuf := make([]byte, 4<<10+rand.Intn(8<<10))
		atomic.AddInt64(&bytesCreated, int64(len(newBuf))) // avoid optimizing the prior line out
		atomic.AddInt64(&numCreated, 1)
	}
}

// genUpFrontMem generates a bunch of heap fragmented byte slices and then
// discards them all
func genUpfrontMem() {
	var ticker *time.Ticker
	if *startBatches == 0 {
		*startBatches = 1
	} else {
		ticker = time.NewTicker(time.Second / time.Duration(*bps))
		defer ticker.Stop()
		fmt.Printf("generating %d batches of size %d at a rate of %d bufs per second\n",
			*startBatches, *startBufs, *bps)
	}
	for j := 0; j < *startBatches; j++ {
		upfrontBufs := make([][]byte, 0)
		runtime.GC()
		for i := 0; i < *startBufs; i++ {
			if ticker != nil {
				<-ticker.C
			}
			newBuf := make([]byte, 4<<10+rand.Intn(8<<10))
			upfrontBufs = append(upfrontBufs, newBuf)
			atomic.AddInt64(&numCreated, 1)
		}
		fmt.Printf("discarding batch (len %d)...\n", len(upfrontBufs))
	}
	fmt.Println("done generating fragmented upfront mem")
}

func main() {
	flag.Parse()

	interrupt := make(chan os.Signal, 1)
	signal.Notify(interrupt, os.Interrupt)

	// Setup pprof.
	deadPprof := make(chan error)
	go func() {
		deadPprof <- http.ListenAndServe("localhost"+*pprofPort, nil)
	}()

	// Start tracking memstats.
	go printMemStats()
	go func() {
		// Segment the heap if necessary.
		genUpfrontMem()
		// Attempt to return this memory to the OS.
		runtime.GC()
		debug.FreeOSMemory()
		// Start generating a steady pace of ignored slices.
		go creator()
	}()

	select {
	case <-interrupt:
		fmt.Println("received interrupt, quitting")
	case err := <-deadPprof:
		fmt.Printf("pprof server died: %v, quitting", err)
		os.Exit(1)
	}
}

func printMemStats() {
	fmt.Println("ticking memstats...")
	for range time.Tick(time.Second) {
		doPrintMemStats()
	}
}

var cpuLast float64      // tracks cpu % delta between memstat printings
var lastNumCreated int64 // tracks num bufs created delta between memstat printings

func doPrintMemStats() {
	var mem runtime.MemStats
	runtime.ReadMemStats(&mem)
	fmt.Printf("===\n"+
		"total created:\t%d\n"+
		"since last:\t%d\n"+
		"mem alloc:\t%d\n"+
		"mem sys:\t%d\n"+
		"last gc pause:\t%.6fms\n"+
		"since last gc:\t%.6fms\n",
		atomic.LoadInt64(&numCreated),
		numCreated-lastNumCreated,
		mem.Alloc,
		mem.Sys,
		float64(mem.PauseNs[(mem.NumGC+255)%256])/1e6,
		float64(time.Now().Sub(time.Unix(0, int64(mem.LastGC))).Nanoseconds())/1e6)
	lastNumCreated = numCreated
	var r syscall.Rusage
	if syscall.Getrusage(syscall.RUSAGE_SELF, &r) == nil {
		cpu := 100 * float64((r.Utime.Nano() + r.Stime.Nano())) / 1e9
		fmt.Printf("cpu pct:\t%.4f%%\n", cpu-cpuLast)
		cpuLast = cpu
	}
	fmt.Println()
}
